/******************************************************************************

   scnsm.c
   Copyright (c) 1996-1998, Soundcode, Inc. All rights reserved.

   This code and information is provided "as is" without warranty 
   of any kind, either expressed or implied, including but not 
   limited to the implied warranties of merchantability and/or 
   fitness for a particular purpose. 

   08/15/98 -- Baseline, Jim Adler
   11/04/98 -- 1.0 Beta Release

******************************************************************************/

/* System Headers */
#include <basedef.h>
#include <vmm.h>
#include <debug.h>
#include <error.h>
#include <vxdwraps.h>
#include <vwin32.h>

/* Intrinsic Functions */
void* __cdecl memcpy(void*, const void *, int);
#pragma intrinsic(memcpy)

void* __cdecl memset(void*, int, int);
#pragma intrinsic(memset)

/* Public Headers */
#include "scnsm.h"

/* Private Headers */
#include "scwraps.h"

/* Constants */
#define NSM_MAJOR_VERSION   1    /* must match defs in scnsmctl.asm */
#define NSM_MINOR_VERSION   0

/* Pragmas */
#pragma warning(disable : 4514)  /* unreferenced inline function has been removed */
#pragma warning(disable : 4100)  /* unreferenced formal parameter */

/* Macros */
#ifndef MAKELONG
   #define MAKELONG(low, high) ((LONG)(((WORD)(low)) | (((DWORD)((WORD)(high))) << 16)))
#endif
#ifndef MAKELP
   #define MAKELP(sel, off)    ((void*)MAKELONG((off), (sel)))
#endif

#if DEBLEVEL > DEBLEVELRETAIL
char g_pDbg[256];
#define TRACE(str)	Out_Debug_String(str)
#define TRACEVAR(var, type)   {_Sprintf(g_pDbg, #var" = %"#type"\r\n", var); TRACE(g_pDbg);}
#else
#define TRACE(str)
#define TRACEVAR(var, type)
#endif

#ifdef DEBUG
   #define ASSERT(e)\
      if(!(e))\
      {\
         _Sprintf(g_pDbg, "Assertion failure in " __FILE__ ", line %d: " #e "\r\n", __LINE__);\
         Out_Debug_String(g_pDbg);\
      }
#else
	#define ASSERT(e)
#endif

/* Functions */
static WORD GetVersion();
static DWORD GetPageSize();
static void* AllocPage(DWORD nPages, DWORD type, DWORD* phMem);
static DWORD AllocPagePm(WORD nPages, void** ppBufPm, void** ppBuf, DWORD* phBuf, WORD wVersion);
static DWORD MapPmPage(WORD nPages, void* pBuf, void** ppBufPm);
static DWORD FreePage(void* p);

/* Globals */
static BOOL g_bLocked = FALSE;
static WORD g_nPages = 1;
static void* g_pBuf = 0;
static DWORD g_hBuf = 0;
static void* g_pBufPm = 0;

/*-----------------*/
static WORD GetVersion()
{
   return Get_VMM_Version();
}

/*-----------------*/
static DWORD GetPageSize()
{
   return PAGESIZE;
}

static void* AllocPage(DWORD nPages, DWORD type, DWORD* phMem)
{
   void* p = 0;
   DWORD h = 0;
   HVM hVM = type == PG_SYS ? 0 : Get_Cur_VM_Handle();
   DWORD minPage = 0x0L;
   DWORD maxPage = 0x10000L;   // 256 MB
   BOOL bSuccess = 
      PageAllocateX(
         nPages,                 /* size in pages */
         type,                   /* allocate from ring-3 (PG_VM) or ring-0 (PG_SYS) area */
         hVM,                    /* must be 0 for PG_SYS */
         0,                      /* alignment 0 -> PAGESIZE alignment */
         minPage,                /* min physical page */
         maxPage,                /* max physical page */
         0,                      /* return physical address */
         PAGECONTIG | PAGEUSEALIGN | PAGEFIXED,
         &h,
         &p);
   if( phMem )
      *phMem = h;
   return p;
}

static DWORD AllocPagePm(WORD nPages, void** ppBufPm, void** ppBuf, DWORD* phBuf, WORD wVersion)
{
   DWORD rc = ERROR_NOT_ENOUGH_MEMORY;
   void* pBuf = 0;

   /* Parameter check */
   if( !nPages || !ppBufPm || !ppBuf || !phBuf )
      return ERROR_INVALID_PARAMETER;

   /* Default outputs */
   *ppBufPm = 0;
   *ppBuf = 0;
   *phBuf = 0;

   /* Win3.x just gets the global memory address */
   if( HIBYTE(wVersion) <= 3 )
   {
      if( !g_pBuf )
         return ERROR_NOT_ENOUGH_MEMORY;
      if( !g_pBufPm )
      {
         DWORD _rc = MapPmPage(nPages, g_pBuf, &g_pBufPm);
         if( _rc )
            return _rc;
      }
      if( g_bLocked )
         return ERROR_LOCK_VIOLATION;
      g_bLocked = TRUE;
      if( nPages > g_nPages )
         return rc;
      *ppBuf = g_pBuf;
      *phBuf = g_hBuf;
      *ppBufPm = g_pBufPm;
      return NO_ERROR;
   }

   /* Allocate */
   TRACEVAR(nPages, ld);
   pBuf = AllocPage(nPages, PG_VM, phBuf);
   TRACEVAR(pBuf, lx);
   if( !pBuf )
      return rc;
   *ppBuf = pBuf;
   TRACEVAR(*ppBuf, lx);

   /* Map it */
   rc = MapPmPage(nPages, pBuf, ppBufPm);
   return rc;
}

static DWORD MapPmPage(WORD nPages, void* pBuf, void** ppBufPm)
{
   DWORD rc = NO_ERROR;
   DWORD segment = 0;
   DWORD offset = 0;
   PVMMCB hVM = (PVMMCB)Get_Cur_VM_Handle();

   if( !ppBufPm )
      return ERROR_INVALID_PARAMETER;

   /* Map_Lin_To_VM_Addr is good only in System VM */
   *ppBufPm = 0;
   if( !Test_Sys_VM_Handle(hVM) ||
       !(hVM->CB_VM_Status & VMSTAT_PM_EXEC) ||
       !Map_Lin_To_VM_Addr(pBuf, nPages*PAGESIZE-1, &segment, &offset) )
      return ERROR_BAD_ENVIRONMENT;
   *ppBufPm = MAKELP(segment, offset);
   TRACEVAR(*(DWORD*)ppBufPm, lx);
   if( *(DWORD*)ppBufPm )
      rc = NO_ERROR;
   return rc;
}

static DWORD FreePage(void* p)
{
   if( !p )
      return NO_ERROR;
   return !_PageFree(p, 0);
}

/*-----------------*/
BOOL OnInitComplete(DWORD hVM)
{
   TRACE("OnInitComplete\r\n");

   /* Win3.x doesn't allow allocation of fixed, contiguous pages dynamically.
    * So, allocate a page here and use it for all requests.
    *
    * BUGBUG - this sucks! At least, should allocate a configurable set of pages
    *          and dole them out on a per VM basis.
    */
   if( HIBYTE(GetVersion()) <= 3 )
      AllocPagePm(g_nPages, &g_pBufPm, &g_pBuf, &g_hBuf, (WORD)-1);
   return TRUE;
}

BOOL OnSystemExit(DWORD hVM)
{
   TRACE("OnSystemExit\r\n");
   if( g_hBuf )
      FreePage((void*)g_hBuf);
   return TRUE;
}

/*-----------------*/
BOOL OnSysDynamicDeviceInit(DWORD hVM)
{
   TRACE("OnSysDynamicDeviceInit\r\n");
   return TRUE;
}

BOOL OnSysDynamicDeviceExit(DWORD hVM)
{
   TRACE("OnSysDynamicDeviceExit\r\n");
   return TRUE;
}

/*-----------------*/
DWORD OnW32DeviceIoControl(PDIOCPARAMETERS p)
{
   /*
   DeviceIoControl            DIOCPARAMETERS
   ===============            ==============
   DWORD dwIoControlCode      DWORD	dwIoControlCode
   LPVOID lpInBuffer          DWORD	lpvInBuffer
   DWORD nInBufferSize        DWORD	cbInBuffer
   LPVOID lpOutBuffer         DWORD	lpvOutBuffer
   DWORD nOutBufferSize       DWORD	cbOutBuffer
   LPDWORD lpBytesReturned    DWORD	lpcbBytesReturned
   LPOVERLAPPED lpOverlapped  DWORD	lpoOverlapped
   */

   DWORD rc = ERROR_GEN_FAILURE;

   switch( p->dwIoControlCode )
   {
      case DIOC_CLOSEHANDLE:
         TRACE("OnW32DeviceIoControl::CLOSEHANDLE\r\n");
         rc = NO_ERROR;
         break;

      case DIOC_OPEN:
         TRACE("OnW32DeviceIoControl::OPEN\r\n");
         rc = NO_ERROR;
         break;

      case NSM_GETVERSION:
      {
         TRACE("OnW32DeviceIoControl::GETVERSION\r\n");
         if( !p->lpvOutBuffer || p->cbOutBuffer < sizeof(LONG) )
         {
            rc = ERROR_INVALID_PARAMETER;
            break;
         }
         *(DWORD*)p->lpvOutBuffer = MAKELONG(NSM_MINOR_VERSION, NSM_MAJOR_VERSION);
         if( p->lpcbBytesReturned )
            *(DWORD*)p->lpcbBytesReturned = sizeof(LONG);
         rc = NO_ERROR;
         break;
      }

      case NSM_GETPAGESIZE:
      {
         TRACE("OnW32DeviceIoControl::NSM_GETPAGESIZE\r\n");
         if( !p->lpvOutBuffer || p->cbOutBuffer < sizeof(LONG) )
         {
            rc = ERROR_INVALID_PARAMETER;
            break;
         }
         *(DWORD*)p->lpvOutBuffer = GetPageSize();
         if( p->lpcbBytesReturned )
            *(DWORD*)p->lpcbBytesReturned = sizeof(LONG);
         rc = NO_ERROR;
         break;
      }

      case NSM_ALLOCPAGE:
      {
         TRACE("OnW32DeviceIoControl::NSM_ALLOCPAGE\r\n");
         rc = ERROR_NOT_ENOUGH_MEMORY;
         if( !p->lpvOutBuffer || !p->cbOutBuffer )
         {
            rc = ERROR_INVALID_PARAMETER;
            break;
         }
         *(BYTE**)p->lpvOutBuffer = AllocPage(p->cbOutBuffer, PG_VM, 0);
         TRACEVAR(*(BYTE**)p->lpvOutBuffer, x);
         if( *(BYTE**)p->lpvOutBuffer )
            rc = NO_ERROR;
         break;
      }
   
      case NSM_FREEPAGE:
      {
         TRACE("OnW32DeviceIoControl::NSM_FREEPAGE\r\n");
         if( !p->lpvInBuffer )
         {
            rc = NO_ERROR;
            break;
         }
      	if( !FreePage((void*)p->lpvInBuffer) )
            rc = NO_ERROR;
         break;
      }

      default:
         rc = ERROR_INVALID_FUNCTION;
         break;
   }
   return rc;
}

/*-----------------*/
DWORD _stdcall OnPMApi(DWORD dwCode, DWORD dwParm1, DWORD dwParm2, DWORD dwParm3)
{
   DWORD rc = ERROR_GEN_FAILURE;

   switch( (WORD)dwCode )
   {
      case NSM_GETVERSION:
      {
         TRACE("OnPMApi::NSM_GETVERSION\r\n");
         if( !dwParm1 || dwParm2 < sizeof(LONG) )
         {
            rc = ERROR_INVALID_PARAMETER;
            break;
         }
      	dwParm1 = Client_Ptr_Flat(DS, BX);  /* DS:BX holds dwParm1 */
         *(DWORD*)dwParm1 = MAKELONG(NSM_MINOR_VERSION, NSM_MAJOR_VERSION);
         rc = NO_ERROR;
         break;
      }

      case NSM_GETPAGESIZE:
      {
         TRACE("OnPMApi::NSM_GETPAGESIZE\r\n");
         if( !dwParm1 || dwParm2 < sizeof(SHORT) )
         {
            rc = ERROR_INVALID_PARAMETER;
            break;
         }
      	dwParm1 = Client_Ptr_Flat(DS, BX);  /* DS:BX holds dwParm1 */
         *(DWORD*)dwParm1 = PAGESIZE;
         rc = NO_ERROR;
         break;
      }

      case NSM_ALLOCPAGE:
      {
         void* pBuf = 0;
         TRACE("OnPMApi::NSM_ALLOCPAGE\r\n");
         if( !dwParm1 || !dwParm2 || !dwParm3 )
         {
            rc = ERROR_INVALID_PARAMETER;
            break;
         }
      	dwParm2 = Client_Ptr_Flat(DS, CX);  /* DS:CX holds dwParm2, the PM address */
      	dwParm3 = Client_Ptr_Flat(DS, DX);  /* DS:DX holds dwParm3, the mem handle */
         rc = AllocPagePm((WORD)dwParm1, (void**)dwParm2, &pBuf, (DWORD*)dwParm3, GetVersion());
         TRACEVAR(dwParm1, x);
         TRACEVAR(*(DWORD*)dwParm2, x);
         TRACEVAR(*(DWORD*)dwParm3, x);
         break;
      }

      case NSM_FREEPAGE:
      {
         DWORD hBuf;
         TRACE("OnPMApi::NSM_FREEPAGE\r\n");
         if( !dwParm1 )
         {
            rc = NO_ERROR;
            break;
         }
         dwParm1 = Client_Ptr_Flat(DS, BX);  /* DS:BX holds dwParm1, the mem handle */
         hBuf = *(DWORD*)dwParm1;
         if( hBuf == g_hBuf )
            g_bLocked = FALSE, rc = NO_ERROR;
         else
            if( !FreePage((void*)hBuf) )
               rc = NO_ERROR;
         break;
      }

      default:
         rc = ERROR_INVALID_FUNCTION;
         break;
   }
   return rc;
}
